require 'erb'

module Web
  class Session
    attr_reader :hint, :username, :expires
    def initialize
      @producers = {}
      @hint = nil
      @username = nil
      @expires = Time.now + 1 * 24 * 60 * 60 # 1 day
    end
    def service(context)
      request(context)
    end
    def request(context)
      params = context.req_params
      oid = params['oid'][0]
      return unless oid
      prod = fetch(oid)
      prod.do_request(context, params) if prod
    end
    def fetch(oid)
      @producers[oid]
    end
    def entry(prod)
      @producers[prod.oid] = prod
    end
    def expired?
      expires = @expires
      expires && Time.now > expires
    end
  end
  class SessionDB
    def initialize(factory)
      @db = {}
      @factory = factory
    end
    def mk_session(context)
      session = @factory.new
      session
    end
    def fetch(context)
      session_id = context.req_cookie('session')
#     session_id = context.req_params('session')[0]
      session = @db[session_id]
      if session && session.expired?
        @db[session_id] = nil
        session = nil
      end
      session ||= mk_session(context)
      store_session(context, session)
      session
    end
    def store_session(context, session)
      session_id = session.object_id.to_s
      @db[session_id] = session
      expires = session.expires
      context.res_add_cookie('session', session_id, expires)
      hint = session.hint
      if hint
        expires = Time.now + 60 * 24 * 60 * 60 # 60 days
        context.res_add_cookie('hint', hint, expires)
      end
      session_id
    end
    def service(context)
      load 'obj_count.rb'
      session = fetch(context)
      session.service(context)
      context.close
    end
  end
  class Producer
    extend ERB::DefMethod
    def self.set_erb(fname)
      def_erb_method('to_html(context=nil)', fname)
    rescue SyntaxError
      #
    end
    def action(context=nil)
      if context
        String(context.req_script_name) + String(context.req_path_info)
      else
        @action
      end
    end
    def attr(opt)
      ary = opt.collect do |name, value|
        if value
          %Q[#{h(name)}="#{h(value)}"]
        else
          nil
        end
      end.compact
      return nil if ary.size == 0
      ary.join(' ')
    end
    def elem(name, opt={})
      head = ["#{name}", attr(opt)].compact.join(" ")
      if block_given?
        %Q[<#{head}>\n#{yield}\n</#{name}>]
      else
        %Q[<#{head} />]
      end  
    end
    def to_div(context=nil)
      elem('div', {'class'=>class_name, 'id'=>oid}) do
        begin
          to_html(context)
        rescue
          %Q["<p>error! #{h($!)}</p>"]
        end
      end
    end
    alias :to_s :to_div
    def to_html(context=nil)
      ''
    end
    def class_name
      self.class.to_s
    end
    def oid
      self.object_id.to_s
    end
#   attr_reader :session_id
    def initialize(session)
      @session = session
#     @session_id = session.object_id.to_s
      session.entry(self)
      @action = nil
    end
    def do_request(context, params)
      cmd = params['cmd'][0]
      cmd = 'do_' + cmd
      if respond_to?(cmd)
        send cmd, context, params
      else
        do_else context, params
      end
    end
    def form(cmd, params={}, context=nil)
#     params['session'] = session_id
      params['oid'] = oid
      params['cmd'] = cmd
      ary = [
        %Q[<form method="POST" action="#{action(context)}">]
      ] << params.map do |name, value|
        %Q[<input type=hidden name="#{h(name)}" value="#{h(value)}" />]
      end
      ary.join("\n")
    end
    def href(cmd, params={}, context=nil)
#     params['session'] = session_id
      params['oid'] = oid
      params['cmd'] = cmd
      ary = params.map do |name, value|
        %Q[#{u(name)}=#{u(value)}]
      end
      %Q[#{action(context)}?#{ary.join(';')}]
    end
    def a(cmd, params={}, context=nil)
      %Q[<a href="#{href(cmd, params, context)}">]
    end
    def do_else(context, params)
      #
    end
    def h(s)
      s.to_s.
      gsub(/&/n, '&amp;').
      gsub(/"/n, '&quot;').
      gsub(/>/n, '&gt;').
      gsub(/</n, '&lt;')
    end
    def u(s)
      s.to_s.gsub(/([^ a-zA-Z0-9_.-]+)/n) do
        '%' + $1.unpack('H2' * $1.size).join('%').upcase
      end.tr(' ', '+')
    end
  end
  class Application
    def initialize(factory)
      @session_db = SessionDB.new(factory)
    end
    def run(context)
      @session_db.service(context)
    end
  end
end

